// vim:expandtab:autoindent:tabstop=4:shiftwidth=4:filetype=c:
/*
 * lsi_raid_inventory
 *   Copyright (c) 2006 Dell, Inc.
 *   by Matt Domsch <Matt_Domsch@dell.com>
 *   Python module-ization by: Michael Brown <Michael_E_Brown@dell.com>
 *
 * based on LSI megaraid_mbox and megaraid_mm device drivers
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   version 2 as published by the Free Software Foundation.
 */

#include <Python.h>

#include <string.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include "lsi_raid_inventory.h"

#define FC_NEW_CONFIG       0xA1
#define NC_SUBOP_PRODUCT_INFO   0x0E

// The exception for this module
static PyObject *LsiInventoryError;


static int get_adapter_attr(int megadev, int adapno,
                struct mcontroller *mc)
{
    struct uioctl_t ioc;

    memset(&ioc, 0, sizeof(struct uioctl_t));

    ioc.outlen = sizeof(struct mcontroller);
    ioc.data = (__u8 *) mc;
    ioc.ui.fcs.buffer = (__u8 *) mc;

    ioc.ui.fcs.opcode = 0x82;
    ioc.ui.fcs.subopcode = MEGAIOC_QADAPINFO;   /* Get adapter info. */

    ioc.ui.fcs.adapno = MKADAP(adapno); /* Get info for this adapter */

    if (ioctl(megadev, MEGAIOCCMD, &ioc) == -1)
        return -1;

    return 0;
}

static int get_adapter_firmware(int megadev, int adapno,
                struct mraid_pinfo *p)
{
    struct uioctl_t ioc;

    memset(p, 0, sizeof(*p));
    memset(&ioc, 0, sizeof(struct uioctl_t));

    ioc.outlen = sizeof(*p);
    ioc.data = (__u8 *) p;
    ioc.ui.fcs.buffer = (__u8 *) p;

    ioc.ui.fcs.opcode = 0x80;

    ioc.mbox[0] = FC_NEW_CONFIG;
    ioc.mbox[2] = NC_SUBOP_PRODUCT_INFO;

    ioc.ui.fcs.adapno = MKADAP(adapno); /* Get info for this adapter */

    if (ioctl(megadev, MEGAIOCCMD, &ioc) == -1)
        return -1;

    return 0;
}

static int get_n_adapters(int megadev)
{
    struct uioctl_t ioc;
    int nadap = 0;

    /*
     * Get the number of adapters as detected by the controller.
     */
    memset(&ioc, 0, sizeof(struct uioctl_t));

    ioc.outlen = sizeof(int);
    ioc.data = (__u8 *) & nadap;

    ioc.ui.fcs.opcode = 0x82;
    ioc.ui.fcs.subopcode = MEGAIOC_QNADAP;  /* Query # of adapters */

    /*
     * Fire the ioctl to get the number of controllers in the system. The
     * return value has the number of adapters, if successful.
     */
    if ((nadap = ioctl(megadev, MEGAIOCCMD, &ioc)) == -1) {
        return -1;
    }
    return nadap;
}


static PyObject *getNumberAdapters( PyObject *self, PyObject *args )
{
    int nadapt = 0;
    int fd;

    const int bufSize = 256;
    char errBuf[bufSize + 1] = {0,};
    char *errstr = 0;

    if ((fd = open("/dev/megadev0", O_RDONLY)) == -1) {
        errstr = strerror(errno);
        snprintf( errBuf, bufSize, "Could not open megaraid device file. The OS Error was: %s", errstr);
        PyErr_SetString(LsiInventoryError, errBuf);
        return 0;
    }
    nadapt = get_n_adapters(fd);
    close(fd);

    return Py_BuildValue("i", nadapt);
};


static PyObject *getAdapterDetails( PyObject *self, PyObject *args )
{
    int fd = 0, i = 0, nadap = 0, ret = 0;
    struct mcontroller mc;
    struct mraid_pinfo pinfo;

    const int bufSize = 256;
    char outBuf[bufSize + 1] = {0,};
    char *errstr = 0;

    PyObject * retObj = 0;

    /* parse arguments and raise exception if incorrect */
    int okay;
    okay = PyArg_ParseTuple(args, "b", &i);
    if (!okay)
        return 0;

    /* clear vars */
    memset(&mc, 0, sizeof(mc));
    memset(&pinfo, 0, sizeof(pinfo));

    if ((fd = open("/dev/megadev0", O_RDONLY)) == -1) {
        errstr = strerror(errno);
        snprintf( outBuf, bufSize, "Could not open megaraid device file. The OS Error was: %s: /dev/megadev0", errstr);
        PyErr_SetString(LsiInventoryError, outBuf);
        goto out;
    }

    nadap = get_n_adapters(fd);
    if (i >= nadap || i < 0){
        PyErr_SetString(LsiInventoryError, "Invalid adapter specified.");
        goto out_close;
    }

    ret = get_adapter_attr(fd, i, &mc);
    if (ret<0) {
        PyErr_SetString(LsiInventoryError, "megadev Get Adapter Info failed\n");
        goto out_close;
    }

    ret = get_adapter_firmware(fd, i, &pinfo);
    if (ret<0) {
        PyErr_SetString(LsiInventoryError, "megadev Get Product Info failed\n");
        goto out_close;
    }

    retObj = Py_BuildValue("{s:(i,i,i,i),s:(i,i,i,i),s:s,s:s}",
        "pci_dbdf", 0, (int)mc.pcibus, (int)mc.pcidev, (int)mc.pcifun,
        "pci_id", mc.pcivendor, mc.pciid, pinfo.subsysvid, pinfo.subsysid,
        "fw_ver", pinfo.fw_version,
        "product", pinfo.product_name
        );

out_close:
    close(fd);

out:
    return retObj;
}



static PyMethodDef moduleMethods[] = {
    { "getNumberAdapters", (PyCFunction) getNumberAdapters, METH_VARARGS, "Returns number of LSI adapters" },
    { "getAdapterDetails", (PyCFunction) getAdapterDetails, METH_VARARGS, "Returns an inventory of LSI adapters. argument is adapter #" },
    { NULL },
};


void initlsi_raid_inventory(void)
{
    PyObject *m, *d;
    m = Py_InitModule("lsi_raid_inventory", moduleMethods);

    d = PyModule_GetDict(m);
    LsiInventoryError = PyErr_NewException("lsi_raid_inventory.LsiInventoryError", NULL, NULL);
    PyDict_SetItemString(d, "LsiInventoryError", LsiInventoryError);
};


