/*
 * lsi_raid_inventory
 *   Copyright (c) 2006 Dell, Inc.
 *   by Matt Domsch <Matt_Domsch@dell.com>
 *
 * based on LSI megaraid_mbox and megaraid_mm device drivers
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   version 2 as published by the Free Software Foundation.
 */

#include <fcntl.h>
#include <linux/ioctl.h>
#include <linux/types.h>
#include <stdint.h>

#define	MAX_CDB_LEN			0x0A
#define MAX_REQ_SENSE_LEN	0x20

struct mega_passthru {
	__u8 timeout:3;
	__u8 ars:1;
	__u8 reserved:3;
	__u8 islogical:1;
	__u8 logdrv;
	__u8 channel;
	__u8 target;
	__u8 queuetag;
	__u8 queueaction;
	__u8 cdb[MAX_CDB_LEN];
	__u8 cdblen;
	__u8 reqsenselen;
	__u8 reqsensearea[MAX_REQ_SENSE_LEN];
	__u8 numsgelements;
	__u8 scsistatus;
	__u32 dataxferaddr;
	__u32 dataxferlen;
};


/*
 * This structure will be passed as argument to each ioctl call to megaraid
 * ioctl node
 */
#pragma pack(1)
struct uioctl_t {
	__u32 inlen;
	__u32 outlen;
	union {
		__u8 fca[16];
		struct {
			__u8 opcode;
			__u8 subopcode;
			__u16 adapno;
#ifdef __LP64__
			__u8 *buffer;
#else
			__u8 *buffer;
			__u8 pad[4];
#endif
			__u32 length;
		} fcs;
	} ui;
	__u8 mbox[18];		/* 16 bytes + 2 status bytes */
	struct mega_passthru pthru;
#ifdef __LP64__
	unsigned char *data;
#else
	unsigned char *data;		/* buffer <= 4096 for 0x80 commands */
	char pad[4];
#endif
};
#pragma pack()

/*
 * The controller attribute structure as returned by the driver
 */
struct mcontroller {
	__u64 base;
	__u8 irq;
	__u8 numldrv;
	__u8 pcibus;
	__u16 pcidev;
	__u8 pcifun;
	__u16 pciid;
	__u16 pcivendor;
	__u8 pcislot;
	__u32 uid;
} __attribute__ ((packed));

/**
 * mraid_pinfo_t - product info, static information about the controller
 * @data_size		: current size in bytes (not including resvd)
 * @config_signature	: Current value is 0x00282008
 * @fw_version		: Firmware version
 * @bios_version	: version of the BIOS
 * @product_name	: Name given to the controller
 * @max_commands	: Maximum concurrent commands supported
 * @nchannels		: Number of SCSI Channels detected
 * @fc_loop_present	: Number of Fibre Loops detected
 * @mem_type		: EDO, FPM, SDRAM etc
 * @signature		:
 * @dram_size		: In terms of MB
 * @subsysid		: device PCI subsystem ID
 * @subsysvid		: device PCI subsystem vendor ID
 * @notify_counters	:
 * @pad1k		: 135 + 889 resvd = 1024 total size
 *
 * This structures holds the information about the controller which is not
 * expected to change dynamically.
 *
 * The current value of config signature is 0x00282008:
 * 0x28 = MAX_LOGICAL_DRIVES,
 * 0x20 = Number of stripes and
 * 0x08 = Number of spans
 */
struct mraid_pinfo {
	uint32_t data_size;
	uint32_t config_signature;
	uint8_t fw_version[16];
	uint8_t bios_version[16];
	uint8_t product_name[80];
	uint8_t max_commands;
	uint8_t nchannels;
	uint8_t fc_loop_present;
	uint8_t mem_type;
	uint32_t signature;
	uint16_t dram_size;
	uint16_t subsysid;
	uint16_t subsysvid;
	uint8_t notify_counters;
	uint8_t pad1k[889];
} __attribute__ ((packed));

/**
 * struct mraid_hba_info - information about the controller
 *
 * @param pci_vendor_id		: PCI vendor id
 * @param pci_device_id		: PCI device id
 * @param subsystem_vendor_id	: PCI subsystem vendor id
 * @param subsystem_device_id	: PCI subsystem device id
 * @param baseport		: base port of hba memory
 * @param pci_bus		: PCI bus
 * @param pci_dev_fn		: PCI device/function values
 * @param irq			: interrupt vector for the device
 *
 * Extended information of 256 bytes about the controller. Align on the single
 * byte boundary so that 32-bit applications can be run on 64-bit platform
 * drivers withoug re-compilation.
 * NOTE: reduce the number of reserved bytes whenever new field are added, so
 * that total size of the structure remains 256 bytes.
 */
struct mraid_hba_info {

	uint16_t pci_vendor_id;
	uint16_t pci_device_id;
	uint16_t subsys_vendor_id;
	uint16_t subsys_device_id;

	uint64_t baseport;
	uint8_t pci_bus;
	uint8_t pci_dev_fn;
	uint8_t pci_slot;
	uint8_t irq;

	uint32_t unique_id;
	uint32_t host_no;

	uint8_t num_ldrv;
} __attribute__ ((aligned(256), packed));


/*
 * Defines for Driver IOCTL interface, Op-code:0x82
 */
#define	MEGAIOC_MAGIC	'm'
#define	MEGAIOCCMD	0xc0046d00	/* Mega IOCTL command */
#define	MEGAIOC_QNADAP		'm'	/* Query # of adapters */
#define	MEGAIOC_QDRVRVER	'e'	/* Query driver version */
#define	MEGAIOC_QADAPINFO	'g'	/* Query adapter information */
#define MKADAP(adapno)		( MEGAIOC_MAGIC << 8 | (adapno) )
#define	GETADAP(mkadap)		( (mkadap) ^ MEGAIOC_MAGIC << 8 )
